#ifdef __linux__
#define _BSD_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#ifndef USE_NETLIB
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#else
#include <netlib.h>
#endif

#include "pktlib.h"
#include "pktbuf.h"

pktbuf_t pktbuf_make_arp(pktbuf_t pktbuf, struct header_param *param)
{
  struct arphdr *arphdr;
  int hdrsize;
  char *sha, *spa, *tha, *tpa;
  in_addr_t ipaddr;

  hdrsize = sizeof(struct arphdr) +
    param->arp.hardware_addr_size * 2 + param->arp.protocol_addr_size * 2;
  if (param->arp.header_size > 0)
    hdrsize = param->arp.header_size;

  arphdr = (struct arphdr *)pktbuf_add_header(pktbuf, hdrsize);
  memset(arphdr, 0, hdrsize);

  arphdr->ar_hrd = htons(param->arp.hardware);
  arphdr->ar_pro = htons(param->arp.protocol);
  arphdr->ar_hln = param->arp.hardware_addr_size;
  arphdr->ar_pln = param->arp.protocol_addr_size;
  arphdr->ar_op = htons(param->arp.operation);

  sha = (char *)(arphdr + 1);
  spa = sha + param->arp.hardware_addr_size;
  tha = spa + param->arp.protocol_addr_size;
  tpa = tha + param->arp.hardware_addr_size;

  if (param->arp.hardware == ARPHRD_ETHER) {
    memcpy(sha, param->arp.sender_hardware_addr, ETHER_ADDR_LEN);
    memcpy(tha, param->arp.target_hardware_addr, ETHER_ADDR_LEN);
  }

  if (param->arp.protocol == ETHERTYPE_IP) {
    ipaddr = htonl(param->arp.sender_protocol_addr);
    memcpy(spa, &ipaddr, sizeof(in_addr_t));
    ipaddr = htonl(param->arp.target_protocol_addr);
    memcpy(tpa, &ipaddr, sizeof(in_addr_t));
  }

  return pktbuf;
}

pktbuf_t pktbuf_make_icmp(pktbuf_t pktbuf, struct header_param *param)
{
  struct icmp *icmphdr;
  int hdrsize;

  hdrsize = sizeof(*icmphdr);
  if (param->icmp.header_size > 0)
    hdrsize = param->icmp.header_size;

  icmphdr = (struct icmp *)pktbuf_add_header(pktbuf, hdrsize);
  memset(icmphdr, 0, hdrsize);

  icmphdr->icmp_type = param->icmp.type;
  icmphdr->icmp_code = param->icmp.code;

  return pktbuf;
}

pktbuf_t pktbuf_make_tcp(pktbuf_t pktbuf, struct header_param *param)
{
  struct tcphdr *tcphdr;
  int hdrsize;

  hdrsize = sizeof(*tcphdr);
  if (param->tcp.header_size > 0)
    hdrsize = param->tcp.header_size;

  tcphdr = (struct tcphdr *)pktbuf_add_header(pktbuf, hdrsize);
  memset(tcphdr, 0, hdrsize);

  tcphdr->th_sport = htons(param->tcp.src_port);
  tcphdr->th_dport = htons(param->tcp.dst_port);
  tcphdr->th_seq = htonl(param->tcp.seq_number);
  tcphdr->th_ack = htonl(param->tcp.ack_number);
  tcphdr->th_off = param->tcp.offset >> 2;
  tcphdr->th_flags = param->tcp.flags;
  tcphdr->th_win = htons(param->tcp.window);

  return pktbuf;
}

pktbuf_t pktbuf_make_tcp_option(pktbuf_t pktbuf, int type, int length, int value)
{
  char *p;
  uint16_t val16;

  p = pktbuf_add_header(pktbuf, length);
  p[0] = type;

  switch (type) {
  case TCPOPT_NOP:
    break;
  case TCPOPT_MAXSEG:
  case TCPOPT_WINDOW:
    p[1] = length;
    switch (length) {
    case 3: p[2] = value; break;
    case 4: val16 = htons(value); memcpy(&p[2], &val16, 2); break;
    default:
      return NULL;
    }
    break;
  default:
    return NULL;
  }

  return pktbuf;
}

pktbuf_t pktbuf_make_udp(pktbuf_t pktbuf, struct header_param *param)
{
  struct udphdr *udphdr;
  int hdrsize, paysize;

  hdrsize = sizeof(*udphdr);
  if (param->udp.header_size > 0)
    hdrsize = param->udp.header_size;

  paysize = pktbuf_get_size(pktbuf);
  if (param->udp.payload_size > 0)
    paysize = param->udp.payload_size;

  udphdr = (struct udphdr *)pktbuf_add_header(pktbuf, hdrsize);
  memset(udphdr, 0, hdrsize);

  udphdr->uh_sport = htons(param->udp.src_port);
  udphdr->uh_dport = htons(param->udp.dst_port);
  udphdr->uh_ulen  = htons(hdrsize + paysize);

  return pktbuf;
}

pktbuf_t pktbuf_make_ip(pktbuf_t pktbuf, struct header_param *param)
{
  struct ip *iphdr;
  int hdrsize, paysize;

  hdrsize = sizeof(struct ip);
  if (param->ip.header_size > 0)
    hdrsize = param->ip.header_size;

  paysize = pktbuf_get_size(pktbuf);
  if (param->ip.payload_size > 0)
    paysize = param->ip.payload_size;

  iphdr = (struct ip *)pktbuf_add_header(pktbuf, hdrsize);
  memset(iphdr, 0, hdrsize);

  iphdr->ip_v   = 4;
  iphdr->ip_hl  = hdrsize >> 2;
  iphdr->ip_len = htons(hdrsize + paysize);
  iphdr->ip_id  = htons(param->ip.id);
  iphdr->ip_ttl = param->ip.ttl;
  iphdr->ip_p   = param->ip.protocol;
  iphdr->ip_src.s_addr = htonl(param->ip.src_ipaddr);
  iphdr->ip_dst.s_addr = htonl(param->ip.dst_ipaddr);

  return pktbuf;
}

struct vlantag {
  uint16_t type;
  uint16_t tag;
} vlantag;

pktbuf_t pktbuf_insert_vlantag(pktbuf_t pktbuf, int vid)
{
  struct ether_header ether_header;
  struct ether_header *ehdr;
  struct vlantag vtag;

  if (!vid)
    return pktbuf;

  memcpy(&ether_header, pktbuf_get_header(pktbuf), ETHER_HDR_LEN);
  ehdr = (struct ether_header *)pktbuf_add_header(pktbuf, sizeof(vtag));

  memcpy(ehdr, &ether_header, ETHER_HDR_LEN);

  vtag.type = htons(ETHERTYPE_VLAN);
  vtag.tag  = htons(vid & 0x0FFF);
  memcpy(&ehdr->ether_type, &vtag, sizeof(vtag));

  return pktbuf;
}

pktbuf_t pktbuf_delete_vlantag(pktbuf_t pktbuf, int *vid)
{
  struct ether_header ether_header;
  struct ether_header *ehdr;
  struct vlantag vtag;

  ehdr = (struct ether_header *)pktbuf_get_header(pktbuf);
  if (ntohs(ehdr->ether_type) != ETHERTYPE_VLAN) {
    if (vid) *vid = 0;
    return pktbuf;
  }

  memcpy(&ether_header, ehdr, ETHER_HDR_LEN);
  memcpy(&vtag, &ehdr->ether_type, sizeof(vtag));

  if (vid) *vid = ntohs(vtag.tag) & 0x0FFF;

  ehdr = (struct ether_header *)pktbuf_delete_header(pktbuf, sizeof(vtag));
  ether_header.ether_type = ehdr->ether_type;

  memcpy(ehdr, &ether_header, ETHER_HDR_LEN);

  return pktbuf;
}

pktbuf_t pktbuf_make_ethernet(pktbuf_t pktbuf, struct header_param *param)
{
  struct ether_header *ehdr;
  int hdrsize;

  hdrsize = ETHER_HDR_LEN;
  if (param->ethernet.vid)
    hdrsize += sizeof(struct vlantag);
  if (param->ethernet.header_size > 0)
    hdrsize = param->ethernet.header_size;

  ehdr = (struct ether_header *)pktbuf_add_header(pktbuf, hdrsize);
  memset(ehdr, 0, hdrsize);

  memcpy(ehdr->ether_dhost, param->ethernet.dst_macaddr, ETHER_ADDR_LEN);
  memcpy(ehdr->ether_shost, param->ethernet.src_macaddr, ETHER_ADDR_LEN);
  ehdr->ether_type = htons(param->ethernet.type);

  if (hdrsize == ETHER_HDR_LEN + sizeof(struct vlantag)) {
    ehdr->ether_type = htons(ETHERTYPE_VLAN);
    pktbuf_delete_vlantag(pktbuf, NULL);
    pktbuf_insert_vlantag(pktbuf, param->ethernet.vid);
  }

  return pktbuf;
}

int pktbuf_analyze_arp(pktbuf_t pktbuf, struct header_param *param)
{
  struct arphdr *arphdr;
  int r = 0, hdrsize, paysize;
  char *sha, *spa, *tha, *tpa;
  in_addr_t ipaddr;

  if (pktbuf_get_size(pktbuf) < sizeof(struct arphdr))
    return r;

  arphdr = (struct arphdr *)pktbuf_get_header(pktbuf);

  hdrsize = sizeof(struct arphdr) + arphdr->ar_hln * 2 + arphdr->ar_pln * 2;
  if (pktbuf_get_size(pktbuf) < hdrsize)
    return r;

  paysize = pktbuf_get_size(pktbuf) - hdrsize;

  param->arp.header_size = hdrsize;
  param->arp.payload_size = paysize;
  param->arp.hardware = ntohs(arphdr->ar_hrd);
  param->arp.protocol = ntohs(arphdr->ar_pro);
  param->arp.hardware_addr_size = arphdr->ar_hln;
  param->arp.protocol_addr_size = arphdr->ar_pln;
  param->arp.operation = ntohs(arphdr->ar_op);

  sha = (char *)(arphdr + 1);
  spa = sha + param->arp.hardware_addr_size;
  tha = spa + param->arp.protocol_addr_size;
  tpa = tha + param->arp.hardware_addr_size;

  if (param->arp.hardware == ARPHRD_ETHER) {
    memcpy(param->arp.sender_hardware_addr, sha, ETHER_ADDR_LEN);
    memcpy(param->arp.target_hardware_addr, tha, ETHER_ADDR_LEN);
  }

  if (param->arp.protocol == ETHERTYPE_IP) {
    memcpy(&ipaddr, spa, sizeof(in_addr_t));
    param->arp.sender_protocol_addr = ntohl(ipaddr);
    memcpy(&ipaddr, tpa, sizeof(in_addr_t));
    param->arp.target_protocol_addr = ntohl(ipaddr);
  }

  r++;

  return r;
}

int pktbuf_analyze_icmp(pktbuf_t pktbuf, struct header_param *param)
{
  struct icmp *icmphdr;
  int r = 0, hdrsize, paysize;

  if (pktbuf_get_size(pktbuf) < sizeof(*icmphdr))
    return r;

  icmphdr = (struct icmp *)pktbuf_get_header(pktbuf);

  hdrsize = sizeof(*icmphdr);
  paysize = pktbuf_get_size(pktbuf) - hdrsize;

  param->icmp.header_size = hdrsize;
  param->icmp.payload_size = paysize;
  param->icmp.type = icmphdr->icmp_type;
  param->icmp.code = icmphdr->icmp_code;
  r++;

  return r;
}

int pktbuf_analyze_tcp(pktbuf_t pktbuf, struct header_param *param)
{
  struct tcphdr *tcphdr;
  int r = 0, hdrsize, paysize;

  if (pktbuf_get_size(pktbuf) < sizeof(*tcphdr))
    return r;

  tcphdr = (struct tcphdr *)pktbuf_get_header(pktbuf);

  hdrsize = sizeof(*tcphdr);
  paysize = pktbuf_get_size(pktbuf) - hdrsize;

  param->tcp.header_size = hdrsize;
  param->tcp.payload_size = paysize;
  param->tcp.src_port = ntohs(tcphdr->th_sport);
  param->tcp.dst_port = ntohs(tcphdr->th_dport);
  param->tcp.seq_number = ntohl(tcphdr->th_seq);
  param->tcp.ack_number = ntohl(tcphdr->th_ack);
  param->tcp.offset = tcphdr->th_off << 2;
  param->tcp.flags  = tcphdr->th_flags;
  param->tcp.window = ntohs(tcphdr->th_win);
  r++;

  return r;
}

int pktbuf_analyze_udp(pktbuf_t pktbuf, struct header_param *param)
{
  struct udphdr *udphdr;
  int r = 0, hdrsize, paysize;

  if (pktbuf_get_size(pktbuf) < sizeof(*udphdr))
    return r;

  udphdr = (struct udphdr *)pktbuf_get_header(pktbuf);

  hdrsize = sizeof(*udphdr);

  if (pktbuf_get_size(pktbuf) < ntohs(udphdr->uh_ulen))
    return r;
  paysize = ntohs(udphdr->uh_ulen) - hdrsize;

  param->udp.header_size = hdrsize;
  param->udp.payload_size = paysize;
  param->udp.src_port = ntohs(udphdr->uh_sport);
  param->udp.dst_port = ntohs(udphdr->uh_dport);
  r++;

  return r;
}

int pktbuf_analyze_ip(pktbuf_t pktbuf, struct header_param *param)
{
  struct ip *iphdr;
  int r = 0, hdrsize, paysize;

  if (pktbuf_get_size(pktbuf) < sizeof(struct ip))
    return r;

  iphdr = (struct ip *)pktbuf_get_header(pktbuf);

  if (iphdr->ip_v != 4)
    return r;

  hdrsize = iphdr->ip_hl << 2;
  if (hdrsize < sizeof(struct ip))
    return r;
  if (pktbuf_get_size(pktbuf) < hdrsize)
    return r;

  if (pktbuf_get_size(pktbuf) < ntohs(iphdr->ip_len))
    return r;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  param->ip.header_size = hdrsize;
  param->ip.payload_size = paysize;
  param->ip.id = ntohs(iphdr->ip_id);
  param->ip.ttl = iphdr->ip_ttl;
  param->ip.protocol = iphdr->ip_p;
  param->ip.src_ipaddr = ntohl(iphdr->ip_src.s_addr);
  param->ip.dst_ipaddr = ntohl(iphdr->ip_dst.s_addr);
  r++;

  pktbuf_delete_header(pktbuf, hdrsize);

  switch (param->ip.protocol) {
  case IPPROTO_ICMP:
    r += pktbuf_analyze_icmp(pktbuf, param);
    break;
  case IPPROTO_TCP:
    r += pktbuf_analyze_tcp(pktbuf, param);
    break;
  case IPPROTO_UDP:
    r += pktbuf_analyze_udp(pktbuf, param);
    break;
  default:
    break;
  }

  pktbuf_add_header(pktbuf, hdrsize);

  return r;
}

int pktbuf_analyze_ethernet(pktbuf_t pktbuf, struct header_param *param)
{
  struct ether_header *ehdr;
  int r = 0, hdrsize, paysize;

  if (pktbuf_get_size(pktbuf) < ETHER_HDR_LEN)
    return r;

  ehdr = (struct ether_header *)pktbuf_get_header(pktbuf);

  hdrsize = ETHER_HDR_LEN;
  if (ntohs(ehdr->ether_type) == ETHERTYPE_VLAN)
    hdrsize += sizeof(struct vlantag);
  if (pktbuf_get_size(pktbuf) < hdrsize)
    return r;

  paysize = pktbuf_get_size(pktbuf) - hdrsize;

  param->ethernet.header_size = hdrsize;
  param->ethernet.payload_size = paysize;
  memcpy(param->ethernet.dst_macaddr, ehdr->ether_dhost, ETHER_ADDR_LEN);
  memcpy(param->ethernet.src_macaddr, ehdr->ether_shost, ETHER_ADDR_LEN);
  param->ethernet.type = ntohs(ehdr->ether_type);
  r++;

  pktbuf_delete_vlantag(pktbuf, &param->ethernet.vid);

  pktbuf_delete_header(pktbuf, ETHER_HDR_LEN);

  switch (param->ethernet.type) {
  case ETHERTYPE_ARP:
    r += pktbuf_analyze_arp(pktbuf, param);
    break;
  case ETHERTYPE_IP:
    r += pktbuf_analyze_ip(pktbuf, param);
    break;
  default:
    break;
  }

  pktbuf_add_header(pktbuf, ETHER_HDR_LEN);

  pktbuf_insert_vlantag(pktbuf, param->ethernet.vid);

  return r;
}
